from typing import List, Optional
from github import Github, Auth
from gru.agents.tools.core.code_loader.base import CodeLoader


class GitHubLoader(CodeLoader):
    """Implementation of CodeLoader for GitHub repositories."""

    def __init__(self, token: str):
        """
        Initialize with GitHub token.

        Args:
            token: GitHub authentication token
        """
        self.token = token

    def _get_repo(self, identifier: str):
        """
        Get a GitHub repository object.

        Args:
            identifier: Repository name in format "owner/repo"

        Returns:
            A GitHub repository object
        """
        auth = Auth.Token(self.token)
        g = Github(auth=auth)
        return g.get_repo(identifier)

    def retrieve_code(self, identifier: str, path: Optional[str] = None) -> str:
        """
        Retrieve code from a GitHub repository.

        Args:
            identifier: Repository name in format "owner/repo"
            path: File path within the repository

        Returns:
            The source code as a string
        """

        repo = self._get_repo(identifier)

        contents = repo.get_contents(path)
        return contents.decoded_content.decode()

    def list_contents(self, identifier: str, path: Optional[str] = "") -> List[str]:
        """
        List all file paths in a GitHub repository.

        Args:
            identifier: Repository name in format "owner/repo"
            path: Optional directory path within the repository to start listing from

        Returns:
            List of file paths in the repository
        """

        repo = self._get_repo(identifier)
        contents = repo.get_contents(path or "")

        result = []
        while contents:
            file_content = contents.pop(0)
            if file_content.type == "dir":
                contents.extend(repo.get_contents(file_content.path))
            else:
                result.append(file_content.path)

        return result


from gru.agents.tools.core.code_generator.base import CodeGenerator
from gru.agents.tools.core.code_generator.models import QueryIntent, RetrievalResult
from gru.agents.prompts.sql import (
    SQL_GENERATOR_USER_PROMPT_TEMPLATE,
    SQL_GENERATOR_SYSTEM_PROMPT,
)


class SQLCodeGenerator(CodeGenerator):
    def __init__(self, llm_client, system_prompt=None, user_prompt_template=None):
        """
        Initializes the SQL Code Generator with optional custom prompts.

        :param llm_client: The LLM client to use for generation.
        :param system_prompt: Optional custom system prompt (default: SQL_GENERATOR_SYSTEM_PROMPT).
        :param user_prompt_template: Optional custom user prompt template (default: SQL_GENERATOR_USER_PROMPT_TEMPLATE).
        """
        self.llm_client = llm_client
        self.system_prompt = system_prompt or SQL_GENERATOR_SYSTEM_PROMPT
        self.user_prompt_template = user_prompt_template or SQL_GENERATOR_USER_PROMPT_TEMPLATE

    async def generate_code(
        self, query: str, intent: QueryIntent, context: RetrievalResult
    ) -> str:
        user_prompt = self.user_prompt_template.format(
            query=query,
            schemas=context.schemas,
            examples=context.examples,
            documentation=context.documentation,
        )
        return await self.llm_client.generate(self.system_prompt, user_prompt)



from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional
from gru.agents.tools.core.vector_db.models import VectorDBConfig


class VectorDBClient(ABC):
    @abstractmethod
    def connect(self, config: VectorDBConfig) -> None:
        """Establishes a connection to the vector database."""
        pass

    @abstractmethod
    async def similarity_search(
        self,
        collection_name: str,
        query_vector: List[float],
        anns_field: str,
        top_k: int = 5,
        search_params: Optional[Dict[str, Any]] = None,
        output_fields: Optional[List[str]] = None,
    ) -> List[Dict[str, Any]]:
        """
        Performs a similarity search using the query and returns top-k results.
        """
        pass

    @abstractmethod
    async def filtered_search(
        self,
        collection_name: str,
        filter_expr: str,
        limit: int = 5,
        output_fields: Optional[List[str]] = None,
    ) -> List[Dict[str, Any]]:
        """Performs a filtered search based on additional query parameters."""
        pass

    @abstractmethod
    async def add_to_collection(
        self, collection_name: str, documents: List[Dict[str, Any]]
    ):
        """Adds new documents to the vector database collection."""
        pass

    @abstractmethod
    async def update_collection(
        self,
        collection_name: str,
        document_id: str,
        update_data: Dict[str, Any]
    ):
        """Updates an existing document in the specified collection."""
        pass

    @abstractmethod
    async def delete_from_collection(
        self,
        collection_name: str,
        document_id: str
    ):
        """Deletes a document from the specified collection."""
        pass

    @abstractmethod
    async def list_collections(self) -> List[str]:
        """
        Lists all available collections in the vector database.
        """
        pass

    @abstractmethod
    async def get_collection_stats(
        self,
        collection_name: str
    ) -> Dict[str, Any]:
        """Returns statistics for a given collection."""
        pass

    @abstractmethod
    async def reset_collection(self, collection_name: str):
        """Clears all data from the specified collection."""
        pass



from pymilvus import connections, Collection
from typing import List, Dict, Any, Optional
from gru.agents.tools.core.vector_db.models import VectorDBConfig
from .base import VectorDBClient


class MilvusClient(VectorDBClient):
    def __init__(self):
        self.connection = None
        self.config = None

    def connect(self, config: VectorDBConfig) -> None:
        """Establishes a connection to the Milvus vector database."""
        connections.connect(
            alias="default",
            host=config.host,
            port=config.port
        )
        self.config = config  # Store the config for later use
        print("Connected to Milvus.")

    async def similarity_search(
        self,
        collection_name: str,
        query_vector: List[float],
        anns_field: str,
        top_k: int = 5,
        search_params: Optional[Dict[str, Any]] = None,
        output_fields: Optional[List[str]] = None,
    ) -> List[Dict[str, Any]]:
        """
        Performs a similarity search using the query vector
        and returns top-k results.
        """
        collection = Collection(collection_name)

        if search_params is None:
            search_params = {
                "metric_type": self.config.metric if self.config else "L2",
                "params": {
                    "nprobe": self.config.nprobe if self.config else 10
                },
            }

        # Perform the search
        results = collection.search(
            data=[query_vector],
            anns_field=anns_field,
            param=search_params,
            limit=top_k,
            output_fields=output_fields,
        )

        # Process and format the results
        search_results = []
        for hits in results:
            for hit in hits:
                result = {"score": hit.score}
                if output_fields:
                    for field in output_fields:
                        result[field] = hit.entity.get(field)
                search_results.append(result)

        return search_results

    async def filtered_search(
        self,
        collection_name: str,
        filter_expr: str,
        limit: int = 5,
        output_fields: Optional[List[str]] = None,
    ) -> List[Dict[str, Any]]:
        """Performs a filtered search based on additional query parameters."""
        collection = Collection(collection_name)

        results = collection.query(
            expr=filter_expr, output_fields=output_fields, limit=limit
        )

        return results

    async def add_to_collection(
        self, collection_name: str, documents: List[Dict[str, Any]]
    ):
        raise NotImplementedError("Method not implemented yet")

    async def update_collection(
        self,
        collection_name: str,
        document_id: str,
        update_data: Dict[str, Any]
    ):
        raise NotImplementedError("Method not implemented yet")

    async def delete_from_collection(
        self,
        collection_name: str,
        document_id: str
    ):
        raise NotImplementedError("Method not implemented yet")

    async def list_collections(self) -> List[str]:
        raise NotImplementedError("Method not implemented yet")

    async def get_collection_stats(
        self,
        collection_name: str
    ) -> Dict[str, Any]:
        raise NotImplementedError("Method not implemented yet")

    async def reset_collection(self, collection_name: str):
        raise NotImplementedError("Method not implemented yet")



from contextlib import asynccontextmanager
import json
import os
from fastapi import BackgroundTasks, FastAPI, Response
import uvicorn
from gru.agents.checkpoint.task_results import TaskResultsRepository, TaskStatus
from gru.agents.framework_wrappers import AgentWorkflow
from gru.agents.schemas import AgentInvokeRequest, AgentInvokeResponse
from gru.agents.schemas.schemas import AgentConversationRequest, TaskCompleteRequest
import logging
from gru.agents.utils.logging import get_log_fields

agent_name = os.getenv("AGENT_NAME")

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@asynccontextmanager
async def lifespan(app: FastAPI):

    workflow: AgentWorkflow = app.state.workflow
    await workflow.setup()
    app.state.task_results_repo = TaskResultsRepository()
    await app.state.task_results_repo.setup()
    yield


api = FastAPI(lifespan=lifespan)


async def invoke_workflow(request: AgentInvokeRequest):
    try:
        workflow: AgentWorkflow = api.state.workflow
        output = await workflow.invoke(request)
        # Todo: Save output to DB table
        print(output)
    except Exception as e:
        logger.error(
            f"AI agent: invoke api failed - {e}",
            extra=get_log_fields(correlation_id=request.prompt_id),
        )


async def resume_workflow(request: TaskCompleteRequest):
    try:
        workflow: AgentWorkflow = api.state.workflow
        output = await workflow.resume(request)
        # Todo: Save output to DB table
        print(output)
    except Exception as e:
        logger.error(
            f"AI agent: resume workflow failed: {e}",
            extra=get_log_fields(correlation_id=request.prompt_id),
        )


async def update_task_result(request: TaskCompleteRequest):
    try:
        task_results_repo: TaskResultsRepository = api.state.task_results_repo
        await task_results_repo.update(
            agent_name,
            request.prompt_id,
            request.task_type,
            request.tool_call_id,
            TaskStatus.COMPLETED,
            json.dumps(request.result),
        )
    except Exception as e:
        logger.error(
            f"AI agent: Error while upddating task result - {e}",
            extra=get_log_fields(correlation_id=request.prompt_id),
        )


@api.post("/invoke")
async def invoke(
    request: AgentInvokeRequest, background_tasks: BackgroundTasks
) -> AgentInvokeResponse:
    background_tasks.add_task(invoke_workflow, request)
    return AgentInvokeResponse(prompt_id=request.prompt_id)


@api.post("/converse")
async def converse(request: AgentConversationRequest):
    try:
        workflow: AgentWorkflow = api.state.workflow
        return await workflow.converse(request)
    except Exception as e:
        logger.error(
            f"AI agent converse api failed: {e}",
            extra=get_log_fields(correlation_id=request.conversation_id),
        )


@api.post("/task-complete")
async def task_complete(
    request: TaskCompleteRequest, background_tasks: BackgroundTasks
):
    background_tasks.add_task(resume_workflow, request)
    return Response(status_code=200)


@api.post("/save-task-result")
async def save_task_result(
    request: TaskCompleteRequest, background_tasks: BackgroundTasks
):
    background_tasks.add_task(update_task_result, request)
    return Response(status_code=200)


class App:
    def __init__(self, workflow: AgentWorkflow):
        api.state.workflow = workflow

    def run(self):
        uvicorn.run(api, host="0.0.0.0", port=8080)



import json
import os
import uuid
from termcolor import colored
from websockets import Headers
import yaml
from websockets.sync.client import connect

from gru import DEFAULT_CANSO_API_CONFIG
from gru.agents.models import (
    AgentPromptRequest,
    AgentRegisterRequest,
    AgentUpdateRequest,
    MemoryInsertRequest,
    MemoryUpdateRequest,
    MemoryDeleteRequest
)
from gru.schema.api_request_handler import APIRequestHandler
from gru.schema.api_response_handler import APIResponseHandler
from gru.utils.config_reader import ConfigReader

from cookiecutter.main import cookiecutter

AGENT_CONFIG_FILE_NAME = "config.yaml"

AI_AGENT_TEMPLATE_URL = "https://github.com/Yugen-ai/canso-ai-agent-templates.git"





def ai_agent_templates_setup():
    """
    Run Cookiecutter with the specified template.
    """
    cookiecutter(AI_AGENT_TEMPLATE_URL)

def call_api(
        correlation_id:str, 
        auth_token:str, 
        base_url:str, endpoint:str, 
        method:str, 
        body: dict|None = None
    ):
    request_handler = APIRequestHandler(auth_token, base_url=base_url)
    request_handler.set_correlation_id(correlation_id)

    response = request_handler.send_request(method, endpoint, body)

    response_handler = APIResponseHandler(response)
    response_handler.check_for_errors()
    response_message = response_handler.get_message()

    return response_message


def register_agent(
        correlation_id:str, 
        auth_token: str, 
        agent_folder: str, 
        cluster_name: str, 
        image: str, 
        image_pull_secret: str
    ):
    config_file_path = os.path.join(agent_folder, AGENT_CONFIG_FILE_NAME)

    with open(config_file_path, "r") as config_file:
        config_dict = yaml.safe_load(config_file)

    agent_slug = config_dict["agent_name"].lower().replace(' ', '-') 
    
    agent_register_request = AgentRegisterRequest(
        cluster_name=cluster_name,
        agent_name=agent_slug,
        image=image,
        image_pull_secret=image_pull_secret,
        task_server_name=config_dict["task_server_name"],
        checkpoint_db_name=config_dict["checkpoint_db_name"],
        replicas=config_dict["replicas"],
        iam_role_arn=config_dict["iam_role_arn"]
    )

    body = agent_register_request.model_dump()

    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    response_message = call_api(
        correlation_id, auth_token, 
        configs.agents_service_url, configs.agent_register_endpoint, 
        "post", body
    )

    return response_message

def update_agent(
        correlation_id:str, 
        auth_token: str, 
        agent_folder:str, 
        image: str, 
        image_pull_secret: str
    ):
    config_file_path = os.path.join(agent_folder, AGENT_CONFIG_FILE_NAME)

    with open(config_file_path, "r") as config_file:
        config_dict = yaml.safe_load(config_file)

    agent_name = config_dict["agent_name"].lower().replace(' ', '-')
    
    agent_update_request = AgentUpdateRequest(
        image=image,
        image_pull_secret=image_pull_secret,
        task_server_name=config_dict["task_server_name"],
        checkpoint_db_name=config_dict["checkpoint_db_name"],
        replicas=config_dict["replicas"],
        iam_role_arn=config_dict["iam_role_arn"]
    )

    body = agent_update_request.model_dump()

    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    endpoint = configs.agent_update_endpoint.replace("{agent_name}", agent_name)

    response_message = call_api(
        correlation_id, auth_token, 
        configs.agents_service_url, endpoint,
        "patch", body
    )

    return response_message


def deploy_agent(correlation_id:str, auth_token:str, agent_name:str):

    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    endpoint = configs.agent_deploy_endpoint.replace("{agent_name}", agent_name)

    response_message = call_api(
        correlation_id, auth_token, 
        configs.agents_service_url, endpoint, "post"
    )

    return response_message

def delete_agent(correlation_id: str, auth_token: str, agent_name):

    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    endpoint = configs.agent_delete_endpoint.replace("{agent_name}", agent_name)

    response_message = call_api(
        correlation_id, auth_token, 
        configs.agents_service_url, endpoint, "delete"
    )

    return response_message


def prompt_agent(agent_name: str, prompt: dict, auth_token: str):
    """
    Send a prompt to a deployed agent.

    Args:
        agent_name (str): Name of the deployed agent
        prompt (dict): Dictionary containing the prompt data
        auth_token (str): Authentication token for API access

    Returns:
        AgentPromptResponse: Response containing the prompt ID

    Raises:
        ApiError: If the API request fails
        ValueError: If the request parameters are invalid
    """
    prompt_request = AgentPromptRequest(prompt=prompt)
    request_json = prompt_request.model_dump()

    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    request_handler = APIRequestHandler(auth_token, configs)

    endpoint = configs.agent_prompt_endpoint.replace("{agent_name}", agent_name)

    response = request_handler.send_request("post", endpoint, request_json)
    response_handler = APIResponseHandler(response)
    response_handler.check_for_errors()

    response_data = response_handler.get_success_data()
    return response_data


def read_prompt_file(prompt_file: str) -> dict:
    """
    Read and parse a JSON prompt file.

    Args:
        prompt_file (str): Path to the JSON file containing prompt data

    Returns:
        dict: Parsed prompt data

    Raises:
        FileNotFoundError: If the file doesn't exist
        json.JSONDecodeError: If the file contains invalid JSON
    """
    try:
        with open(prompt_file, "r") as f:
            return json.load(f)
    except FileNotFoundError:
        raise FileNotFoundError(f"Prompt file not found: {prompt_file}")
    except json.JSONDecodeError as e:
        raise json.JSONDecodeError(
            f"Invalid JSON in {prompt_file}: {str(e)}", e.doc, e.pos
        )


def converse_agent(
    auth_token: str, agent_name: str, conversation_id: str | None = None
):

    headers = Headers({"authorization": auth_token})

    if conversation_id is None:
        conversation_id = str(uuid.uuid4())[:8]

    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    uri = configs.agent_converse_endpoint.replace("{agent_name}", agent_name).replace(
        "{conversation_id}", conversation_id
    )

    with connect(uri, additional_headers=headers) as websocket:
        print(f"Conversation ID: {conversation_id}")
        while True:
            user_msg = input(colored("User: ", attrs=["bold"]))
            if user_msg == "STOP":
                break

            websocket.send(user_msg)
            response = websocket.recv()
            print(f"{colored('Agent: ', attrs=['bold'])}{response}")

    return "Conversation Stopped"


def memory_insert(
        correlation_id: str,
        auth_token: str,
        agent_name: str,
        file_path: str
) -> str:
    """
    Insert data into agent memory

    Args:
        correlation_id: Unique identifier for the request
        auth_token: Authentication token
        agent_name: Name of the agent
        file_path: Path to JSON file containing memory data

    Returns:
        API response message
    """
    try:
        with open(file_path, 'r') as f:
            memory_data = json.load(f)

        memory_insert_request = MemoryInsertRequest(
            collection_name=memory_data["collection_name"],
            text=memory_data["text"],
            data=memory_data.get("data", {})
        )

        body = memory_insert_request.model_dump()

        configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
        endpoint = configs.agent_memory_endpoint.replace("{agent_name}", agent_name)

        agent_url = f"{configs.agents_service_url}/agents/{agent_name}"

        request_handler = APIRequestHandler(auth_token, base_url=agent_url)
        request_handler.set_correlation_id(correlation_id)

        response = request_handler.send_request("post", endpoint, body)

        response_handler = APIResponseHandler(response)
        response_handler.check_for_errors()

        return response_handler.get_message()
    except FileNotFoundError:
        raise FileNotFoundError(f"Memory file not found: {file_path}")
    except json.JSONDecodeError as e:
        raise json.JSONDecodeError(
            f"Invalid JSON in {file_path}: {str(e)}", e.doc, e.pos
        )


def memory_update(
        correlation_id: str,
        auth_token: str,
        agent_name: str,
        file_path: str
) -> str:
    """
    Update data in agent memory

    Args:
        correlation_id: Unique identifier for the request
        auth_token: Authentication token
        agent_name: Name of the agent
        file_path: Path to JSON file containing update data

    Returns:
        API response message
    """
    try:
        with open(file_path, 'r') as f:
            memory_data = json.load(f)

        memory_update_request = MemoryUpdateRequest(
            collection_name=memory_data["collection_name"],
            memory_id=memory_data["memory_id"],
            text=memory_data["text"],
            data=memory_data.get("data", {})
        )

        body = memory_update_request.model_dump()

        configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
        endpoint = configs.agent_memory_endpoint.replace("{agent_name}", agent_name)
        base_url = f"{configs.agents_service_url}/agents/{agent_name}"

        request_handler = APIRequestHandler(auth_token, base_url=base_url)
        request_handler.set_correlation_id(correlation_id)

        response = request_handler.send_request("patch", endpoint, body)

        response_handler = APIResponseHandler(response)
        response_handler.check_for_errors()

        return response_handler.get_message()
    except FileNotFoundError:
        raise FileNotFoundError(f"Memory file not found: {file_path}")
    except json.JSONDecodeError as e:
        raise json.JSONDecodeError(
            f"Invalid JSON in {file_path}: {str(e)}", e.doc, e.pos
        )


def memory_delete(
        correlation_id: str,
        auth_token: str,
        agent_name: str,
        memory_id: str,
        collection: str
) -> str:
    """
    Delete data from agent memory

    Args:
        correlation_id: Unique identifier for the request
        auth_token: Authentication token
        agent_name: Name of the agent
        memory_id: ID of the memory to delete
        collection: Name of the collection

    Returns:
        API response message
    """

    memory_delete_request = MemoryDeleteRequest(
        collection_name=collection,
        memory_id=memory_id
    )

    body = memory_delete_request.model_dump()

    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    endpoint = configs.agent_memory_endpoint.replace("{agent_name}", agent_name)

    base_url = f"{configs.agents_service_url}/agents/{agent_name}"

    request_handler = APIRequestHandler(auth_token, base_url=base_url)
    request_handler.set_correlation_id(correlation_id)

    response = request_handler.send_request("delete", endpoint, body)

    response_handler = APIResponseHandler(response)
    response_handler.check_for_errors()

    return response_handler.get_message()
